#include <windows.h>
#include <commctrl.h>
#include "halloc.h"
#include "divider.h"
#include "resource.h"

#define BE4(p)			( ((BYTE*)(p))[0] << 24 | ((BYTE*)(p))[1] << 16 | ((BYTE*)(p))[2] << 8 | ((BYTE*)(p))[3] << 0)
#define xisascii(c)		(0x20 <= (c) && (c) <= 0x7E)
#define idivfloor(i, f)	(((i) + (f) - 1) / (f))

struct tagItem{
	BYTE* data;
	DWORD offset;
	DWORD size;
};



void atomRecurse(HWND hTreeView, HTREEITEM parent, BYTE* data, DWORD offset, DWORD length)
{
	DWORD ofs = 0;

	while(ofs < length)
	{
		DWORD size;
		char atom[5];

		size = BE4(data + offset + ofs);
		memcpy(atom, data + offset + ofs + 4, 4);
		atom[4] = '\0';

		if(	!xisascii(atom[0]) ||
			!xisascii(atom[1]) ||
			!xisascii(atom[2]) ||
			!xisascii(atom[3]) )
			break;

		{
			TV_INSERTSTRUCT tvis;
			TCHAR ts[64];
			struct tagItem* it;

			tvis.hParent = parent;
			tvis.hInsertAfter = TVI_LAST;
			tvis.item.mask = TVIF_TEXT | TVIF_PARAM;
			wsprintf(ts, TEXT("%s (+0x%08X:%u[0x%X])"), atom, offset + ofs, size - 8, size - 8);
			tvis.item.pszText = ts;
			it = halloc(sizeof(struct tagItem));
			it->data = data;
			it->offset = offset + ofs + 8;
			it->size = size - 8;
			tvis.item.lParam = (LPARAM)it;

			// ChildNodeSearch
			atomRecurse(hTreeView, TreeView_InsertItem(hTreeView, &tvis), data, offset + ofs + 8, size - 8);
		}

		ofs += size;
	}

	return;
}

void openfile(HWND hTreeView, LPCTSTR fname)
{
	HANDLE hf;
	BYTE* data;
	DWORD size;
	DWORD read;

	if((hf = CreateFile(fname, GENERIC_READ, FILE_SHARE_READ, NULL, OPEN_EXISTING, 0, NULL)) == INVALID_HANDLE_VALUE)
		return;

	TreeView_DeleteAllItems(hTreeView);

	size = GetFileSize(hf, NULL);
	data = halloc(size);
	ReadFile(hf, data, size, &read, NULL);
	// TODO: ErrorCheck
	CloseHandle(hf);

	{
		TV_INSERTSTRUCT tvis;

		tvis.hParent = TVI_ROOT;
		tvis.hInsertAfter = TVI_LAST;
		tvis.item.mask = TVIF_TEXT | TVIF_PARAM;
		tvis.item.pszText = (LPTSTR)fname;
		tvis.item.lParam = 0;

		atomRecurse(hTreeView, TreeView_InsertItem(hTreeView, &tvis), data, 0, size);
	}

	return;
}

/*
void appendWindowText(HWND hWnd, LPCTSTR text)
{
	LPTSTR tmp;

	tmp = halloc(SendMessage(hWnd, WM_GETTEXTLENGTH, 0, 0) + lstrlen(text) + 1);
	GetWindowText(hWnd, tmp, SendMessage(hWnd, WM_GETTEXTLENGTH, 0, 0) + 1);
	lstrcat(tmp, text);
	SetWindowText(hWnd, tmp);

	return;
}
*/

void setDataHex(HWND hEdit, BYTE* data, DWORD offset, DWORD length)
{
	LPTSTR out, os;
	BOOL shift = FALSE;
	UINT startaddr = 0;

	os = out = halloc(77 * (2 + idivfloor(length + (offset & 0x0F), 16)) + 1);

	//				  123456789_123456789_123456789_123456789_123456789_123456789_123456789_1234567
	lstrcpy(os, TEXT("        : +0 +1 +2 +3 +4 +5 +6 +7 +8 +9 +A +B +C +D +E +F :   A S C I I    \r\n")),	os += 77;
	lstrcpy(os, TEXT("--------+-------------------------------------------------+----------------\r\n")),	os += 77;

	startaddr = offset;

	while(length)
	{
		TCHAR ts[256];
		LPTSTR s = ts;
		UINT i, l, o;

		wsprintf(s, TEXT("%08X: "), offset - startaddr),	s += 10;

		if(shift)
			for(i = 0; i < (offset & 0x0F); i++)
				lstrcpy(s, TEXT("   ")),	s += 3;
		else
			i = 0;

		for(l = length, o = offset; l && i < 16; i++, l--)
			wsprintf(s, TEXT("%02X "), *(data + o++)),	s += 3;
		for(; i < 16; i++)
			lstrcpy(s, TEXT("   ")),	s += 3;

		lstrcpy(s, TEXT("|")),	s++;

		if(shift)
			for(i = 0; i < (offset & 0x0F); i++)
				lstrcpy(s, TEXT(" ")),	s ++;
		else
			i = 0;
		for(; length && i < 16; i++, length--, offset++)
			if(0x20 <= *(data + offset) && *(data + offset) <= 0x7E)
				wsprintf(s, TEXT("%c"), *(data + offset)),	s++;
			else
				lstrcpy(s, TEXT(".")),	s++;
		lstrcpy(s, TEXT("\r\n"));
		s += 2;

		lstrcpy(os, ts);
		os += s - ts;
	}

	SetWindowText(hEdit, out);

	hfree(out);

	return;
}



LRESULT CALLBACK MainWndProc(HWND hWnd, UINT msg, WPARAM wParam, LPARAM lParam)
{
	switch(msg)
	{
	case WM_CREATE:
		{
			HWND div = CreateWindowEx(0, WC_DIVIDER, NULL, WS_CHILD | WS_VISIBLE, 0, 0, 0, 0, hWnd, NULL, GetModuleHandle(NULL), (void*)DIVIDER_HORIZONTAL);

			Divider_SetWindow(div, 0, CreateWindowEx(WS_EX_CLIENTEDGE, WC_TREEVIEW, NULL, WS_CHILD | WS_VISIBLE | TVS_HASBUTTONS | TVS_HASLINES | TVS_LINESATROOT | TVS_SHOWSELALWAYS/* | TVS_FULLROWSELECT*/, 0, 0, 0, 0, div, NULL, GetModuleHandle(NULL), NULL));
			Divider_SetWindow(div, 1, CreateWindowEx(WS_EX_CLIENTEDGE, "EDIT", NULL, WS_CHILD | WS_VISIBLE | ES_MULTILINE | WS_VSCROLL | ES_AUTOVSCROLL, 0, 0, 0, 0, div, NULL, GetModuleHandle(NULL), NULL));

			SetWindowLong(hWnd, GWL_USERDATA, (LONG)div);

			{
				LOGFONT lf = {0};

				lf.lfCharSet = DEFAULT_CHARSET;
				lf.lfHeight = 12;
				lstrcpy(lf.lfFaceName, TEXT("lr SVbN"));

				SendMessage(Divider_GetWindow(div, 0), WM_SETFONT, (WPARAM)CreateFontIndirect(&lf), MAKELPARAM(TRUE, 0));
				SendMessage(Divider_GetWindow(div, 1), WM_SETFONT, (WPARAM)CreateFontIndirect(&lf), MAKELPARAM(TRUE, 0));
			}
		}
		return 0;
	case WM_SIZE:
		SetWindowPos((HWND)GetWindowLong(hWnd, GWL_USERDATA), NULL, 0, 0, LOWORD(lParam), HIWORD(lParam), SWP_NOZORDER);
		return 0;
	case WM_DROPFILES:
		{
			TCHAR fn[MAX_PATH];

			DragQueryFile((HDROP)wParam, 0, fn, sizeof(fn));
			DragFinish((HDROP)wParam);

			openfile(Divider_GetWindow((HWND)GetWindowLong(hWnd, GWL_USERDATA), 0), fn);
		}
		return 0;
	case WM_NOTIFY:
		{
			NMHDR* nmh = (void*)lParam;

			if(nmh->code == TVN_DELETEITEM)
			{
				NM_TREEVIEW* nmt = (void*)lParam;

				if(nmt->itemOld.lParam)
					hfree((void*)nmt->itemOld.lParam);
			}
			if(nmh->code == TVN_SELCHANGED)
			{
				NM_TREEVIEW* nmt = (void*)lParam;

				if(nmt->itemNew.lParam)
					setDataHex(Divider_GetWindow((HWND)GetWindowLong(hWnd, GWL_USERDATA), 1),
						((struct tagItem*)nmt->itemNew.lParam)->data,
						((struct tagItem*)nmt->itemNew.lParam)->offset,
						((struct tagItem*)nmt->itemNew.lParam)->size);
				else
					SetWindowText(Divider_GetWindow((HWND)GetWindowLong(hWnd, GWL_USERDATA), 1), NULL);
			}
		}
		break;
	case WM_CLOSE:
		PostQuitMessage(0);
		return 0;
	}

	return DefWindowProc(hWnd, msg, wParam, lParam);
}

HWND initWnd(void)
{
	WNDCLASSEX wc = {0};

	wc.cbSize = sizeof(wc);
	wc.hInstance = GetModuleHandle(NULL);
	wc.hCursor = LoadCursor(NULL, IDC_ARROW);
	wc.hIcon = LoadIcon(wc.hInstance, MAKEINTRESOURCE(IDI_APPICON));
	wc.lpszClassName = TEXT("MainWndProc");
	wc.lpfnWndProc = MainWndProc;

	if(!RegisterClassEx(&wc))
		return NULL;

	return CreateWindowEx(WS_EX_ACCEPTFILES, wc.lpszClassName, TEXT("atomtree - drop me m4v, 3g2, kind of MPEG4..."), WS_THICKFRAME | WS_CAPTION | WS_SYSMENU | WS_MAXIMIZEBOX | WS_MINIMIZEBOX,
		CW_USEDEFAULT, CW_USEDEFAULT, CW_USEDEFAULT, CW_USEDEFAULT,
		NULL, NULL, wc.hInstance, NULL);
}

void WinMainCRTStartup(void)
{
	MSG msg;
	HWND hMainWnd;

	InitCommonControls();
	dividerInit();

	if(!(hMainWnd = initWnd()))
		ExitProcess(0);

	{
		STARTUPINFO si;

		si.cb = sizeof(si);
		GetStartupInfo(&si);
		ShowWindow(hMainWnd, si.wShowWindow);
	}

	while(GetMessage(&msg, NULL, 0, 0))
		DispatchMessage(&msg);

	dividerTerminate();

	ExitProcess(msg.wParam);
}
